package edu.columbia.sip;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/** This class generates and keeps track of SIP Call-IDs for a given SIP Server.
 *  It generates unique Call-IDs and maintains lists of open calls and calls which,
 *  although they may not relate to a currently open connection, are still waiting
 *  for a response or formal closing.<p>
 *
 *  For a definition of Call-IDs see the draft, section 6.12
 */
public class CallIDManager {
    /** maximum size of a Call-ID string */
    public final static int MAXLEN=32;
    /** minimum size of a Call-ID string */
    public final static int MINLEN=5;
    
    /** list to keep track of open calls */
    private Vector _calls;
    
    /** this list keeps track of the requests associated with generated call-ids
     * so that when a response comes in the appropriate action can be taken.
     */
    private Hashtable _callRequests;
    
    /** Default Constructor */
    public CallIDManager() {
        _calls = new Vector();
        _callRequests = new Hashtable();
    }

    /** function which generates a new, unique call ID and adds it to the list of
     *  open calls.
     */
    public String generateCallID() {
        StringBuffer sb; 
        String cid;
        do {
            int length = MINLEN + (int) Math.ceil((MAXLEN - MINLEN) * Math.random());
            sb = new StringBuffer(length);
            for (int i=0; i<length; i++) {
                int d = (int) Math.ceil(36 * Math.random());
                if (d < 36) {
                    sb.append(Character.forDigit(d, Character.MAX_RADIX));
                } else {
                    sb.append('-');
                }
            }
            sb.append('@');
            sb.append(SIPServer.getAddress());
            cid = sb.toString();
        } while (contains(cid));
        
//        System.out.println("CallIDManager.generateCallID(): generated " + cid);
        _calls.addElement(cid);
        return cid;
    }

    /** add this Call-ID (which was taken from an incoming request)
     *  to the open list.  Duplicates are ignored.
     */
    public void add(String callID) {
        _calls.addElement(callID);
    }

    /** remove this Call-ID from the list of active managed IDs. */
    public void remove(String callID) {
        _calls.removeElement(callID);
        _callRequests.remove(callID);
    }

    /** check whether or not the given call-ID is already on the list. */
    public boolean contains(String callID) {
        Enumeration en = _calls.elements();
        while(en.hasMoreElements()) {
            if (en.nextElement().equals(callID)) {
                return true;
            }
        }
        return false;
    } // end contains()
    
    /** associate an existing call-ID with a request message so that a response can 
     * later be correctly handled.
     */
    public void associate(String callID, SIPRequest request) {
        _callRequests.put(callID, request);
    }
    
    /** reverse the effects of associate() */
    public void dissociate(String callID, SIPRequest request) {
        _callRequests.remove(callID);
    }
    
    /** get the associated request for a specified callID, or null if there is no
     * association.
     */
    public SIPRequest getAssociatedRequest(String callID) {
        return (SIPRequest) _callRequests.get(callID);
    }
}