package edu.columbia.sip;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;

import mjs.util.Debug;

/** this class serves two functions.  It keeps track of the list of SIP users who have
 * registered with the server, and it keeps track of the list of SIP users who have
 * expressed interest in the availability information of registrants with this server.
 */
public class RegistrationManager {
    private Vector _availabilityListeners;
    private Hashtable registrantsToLocations;
   
    /** default constructor */
    public RegistrationManager() {
        _availabilityListeners = new Vector();
        registrantsToLocations = new Hashtable();
    }
    
    /** add the supplied availability listener to the list */
    public void addAvailabilityListener(AvailabilityListener al) {
        _availabilityListeners.addElement(al);
    }

    /** remove the supplied availability listener from the list */
    public void removeAvailabilityListener(AvailabilityListener al) {
        _availabilityListeners.removeElement(al);
    }
    
    /** get a list of the availability listeners who are interested 
     * in the availability of the specified SIP user
     */
    public Enumeration getAvailabilityListeners(SIPURL u) {
        Vector v = new Vector();
        Enumeration e = _availabilityListeners.elements();
        while (e.hasMoreElements()) {
            AvailabilityListener al = (AvailabilityListener) e.nextElement();
            if (al.getURL().equals(u)) {
                v.addElement(al);
            }
        }
        return v.elements();
    } // end getAvailabilityListeners()
    
    /** tell all of the availability listeners interested in information about this
     * particular user that he/she has become available
     */
    public void notifyHere(SIPURL u) throws SIPException {
        Debug.println("RegistrationManager.notifyHere(" + u.toString() + ")...");
        for (Enumeration e = getAvailabilityListeners(u); e.hasMoreElements(); ) {
            AvailabilityListener al = (AvailabilityListener) e.nextElement();
            Debug.println("RegistrationManager.notifyHere() notifying a listener of availability.");
            al.here(new AvailabilityEvent(this));
            // if it's a listener from elsewhere, remove it from the list as it's going
            // to send another request and be added right back on.  If it's a local listener
            // it won't be sending another request so it should stay on the list until 
            // it calls removeAL.
            // this is an anomaly because the connection itself sends the automatic GONE or
            // HERE on receiving a response.  If that were left to the listener itself we 
            // wouldn't have to do this here.
            if (al instanceof RemoteAvailabilityListener) {
                removeAvailabilityListener(al);
            }
        }
    } // end notifyHere()

    /** tell all of the availability listeners interested in information about this
     * particular user that he/she has become unavailable.
     */
    public void notifyGone(SIPURL u) throws SIPException {
        Debug.println("RegistrationManager.notifyGone(" + u.toString() + ")...");
        for (Enumeration e = getAvailabilityListeners(u); e.hasMoreElements(); ) {
            AvailabilityListener al = (AvailabilityListener) e.nextElement();
            Debug.println("RegistrationManager.notifyGone() notifying a listener of unavailability.");
            al.gone(new AvailabilityEvent(this));
            // see note above in notifyHere()
            if (al instanceof RemoteAvailabilityListener) {
                removeAvailabilityListener(al);
            }
        }
    } // end notifyGone()

    /** tell the registration manager the the user 'registrant' has requested
     *  to be registered with this SIP Server.  Registrant can be contacted at
     *  the location indicated in the request.
     */
    public void register(SIPURL registrant, String location) {
        registrantsToLocations.put(registrant.toString(), location);
        System.out.println("ResgistrationManager.register(): registered " 
                            + registrant.toString() + " at " + location);                            
    }
    
    /** unregister the (previously registered!) user.*/
    public void unregister(SIPURL registrant, String location) {
        if (registrantsToLocations.containsKey(registrant.toString())) {
            registrantsToLocations.remove(registrant.toString());
            Debug.println("RegistrationManager.unregister() successfully removed registration.");
        } else {
            Debug.println("RegistrationManager.unregister(): " + registrant.toString() + " is not registered here.");
        }
    } // end unregister()
    
    /** check whether a given subject is registered with this server */
    public boolean isRegistered(SIPURL subject) {
        return registrantsToLocations.containsKey(subject.toString());
    }
    
    /** get the location for a given registrant */
    public String getLocationOf(SIPURL subject) {
        return (String) registrantsToLocations.get(subject.toString());
    }

} // end class RegistrationManager   