package edu.columbia.sip;

import java.net.Socket;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;

import mjs.util.Debug;


/** The SIP Parser is responsable for translating messages to and from the SIP
 * standard encoding for SIP servers and user agents.  Each SIP connection
 * has a parser; the SIPConnection passes the socket to the parser.
 *
 * @author Moshe Sambol
 */
public class SIPParser {
    /** The input stream of the connection.*/
    private InputStream _sin;

    /** the reader object connected to the input stream. */
    private BufferedReader _reader;
    
    /** the writer object connected to the output stream. */
    private PrintWriter _writer;
    
    /** The output stream of the connection. */
    private OutputStream _sout;
        
    /** keep the socket for informational purposes */
    private Socket _socket;
    
    /** for debugging purposes only */
    private static int _parserCount = 0;
    private int _thisParserNum = 0;
    
    /** Constructor. 
     *
     * @param i the input stream of the main socket for that server
     * @param o the output stream of the main socket for the server
     *
     * @throws IOException if there is an error working with the socket
     */
    public SIPParser(Socket s) throws IOException {
        _sin = s.getInputStream();
        _reader = new BufferedReader(new InputStreamReader(_sin));
        _sout = s.getOutputStream();
        _writer = new edu.columbia.sip.debug.DebugPrintWriter(_sout);
        _socket = s;
        _thisParserNum = _parserCount++;
        
    }
    
    /** this method can be used to direct a SIPParser to reconnect to a different location */
    public void reconnectTo(Socket s) throws IOException {       
        _socket.close();
        _sin = s.getInputStream();
        _reader = new BufferedReader(new InputStreamReader(_sin));
        _sout = s.getOutputStream();
        _writer = new edu.columbia.sip.debug.DebugPrintWriter(_sout);
        _socket = s;
    }
    
    /** close communication through this parser.  The parser cannot be used again. */
    public void close() throws IOException {
        Debug.println("SIPParser.close() " + _thisParserNum);
        _socket.close();
        _parserCount--;
    }
    
    /** a private method used internally to get a line of data from the socket.
     *
     * @returns a single line of SIP (header or body) data
     */
    private String getLine() throws IOException {
        // get a line from the socket input stream
        String line = _reader.readLine();
   //     Debug.println("SIPParser.getLine() returning a line: " + line);
        // check the line to see if it's legal SIP:
        // is the key a valid token, and is the value a valid 
        // parameterlist for that token?
        // NOT YET DONE // 
        
        // return the line or thrown an exception
        return (line + '\n');
    }
    
    /** a private method used internally to put a line of data to the socket.
     *
     * @param s a line of SIP header or body data
     */
    private void putLine(String s) {
        // check to see if it's legal SIP (same as in getLine())
        // NOT YET DONE
        
        // put it to the socket output stream or throw an exception
        _writer.print(s);
        _writer.flush();
    }

    /** get an entire SIP message as an encapsulated entity.  This includes
     * headers and body.  The parser does minimal checking to see that the message
     * is correctly formatted SIP.  If it it is not an exception is thrown.
     *
     * @returns an encapsulated SIP message, header and body.
     *
     * @throws SIPParseException if there is a poorly formatted line of SIP
     */
    public SIPMessage getMessage() throws SIPParseException, SIPException {
        StringBuffer headers = new StringBuffer();
        StringBuffer contents = new StringBuffer();
        int contentLength = 0;
        String line, identifier;
        SIPMessage msg = null;
        
        // wait for there to be something available on the incoming stream
        try {
            while (! _reader.ready()) {
                Debug.println("SIPParser.getMessage() " + _thisParserNum + " waiting for a message to come in...");
                Thread.sleep(1000);
            }            
        } catch (IOException ioe) {
            Debug.println("SIPMessage.getMessage() got IOE while checking socket's readiness: " 
            + ioe.getMessage());
        } catch (InterruptedException ie) {
            Debug.println("SIPMessage.getMessage() got InterruptedException while trying to sleep: " 
            + ie.getMessage());
        }
        
        StringBuffer debug = new StringBuffer("Getting message ----------------------------------------\n");
        // read one line at a time. find out content length, read content...
        // put all into contents... if can't figure out length throw exception...
        try {
            // get first header line which will distinguish whether this is a
            // request or response, and what version of SIP it represents.
            line = getLine();
            debug.append("got: " + line);
            headers.append(line);
            
            int firstspacepos = line.indexOf(' ');

     //       Debug.println(this, "message header is '" + line + "', firstspacepos is " + firstspacepos);
            
            int type = SIPMessage.getType(line);
            if (type == SIPMessage.REQUEST) {
                String methodstr = line.substring(0, firstspacepos);
                SIPToken method = (SIPToken) SIPToken.getToken(methodstr);
                String urlstr = line.substring(firstspacepos + 1, 
                                        line.indexOf(' ', firstspacepos + 1));
                // pass null as third param since we don't know who it's from yet.
                // the from header will be set later.              
                msg = new SIPRequest(method, new SIPURL(urlstr), null);
            } else if (type == SIPMessage.RESPONSE) {
                // int code, reason string
                int secondspacepos = line.indexOf(' ', firstspacepos + 1);
                String codestr = line.substring(firstspacepos+1, secondspacepos);
                String reason = line.substring(secondspacepos+1, line.indexOf('\n'));
                msg = new SIPResponse(Integer.parseInt(codestr), reason);
            } else {
                throw new SIPException("SIPMessage.getType() returned invalid type: " + type);
            }

            while(true) {
                line = getLine();
                debug.append("got: " + line);
                if (line.equals("\n") || line.length() == 0) {
                    // done with headers, now working on contents.
                    break;
                }
             //   System.out.println("SIPParser.getMessage(): got a non-blank line.");
                int seppos = line.indexOf(SIPToken.SIPSEPARATOR.toString());
                identifier = line.substring(0, seppos);
             //   System.out.println("parsing message.  line is: " + line + " and ident is '"
             //   + identifier + "'");
                if (! SIPToken.isLegalToken(identifier)) {
                    throw new SIPParseException("Illegal SIP token in header: " + line);
                } else {
                    if (identifier.equals(SIPToken.CONTENT_LENGTH.toString())) {
                        String len = line.substring(seppos + 2, line.length() - 1);
                        contentLength = Integer.parseInt(len);
                    }
                } // end if (! SIPToken.isLegalToken(identifier))
                headers.append(line);
            } // end while(reading headers)

            int contentsRead = 0;
            while (contentsRead < contentLength) {
                line = getLine();
                System.out.println("read a line of contents: " + line);
                contentsRead += line.length() + 1;
                contents.append(line);
            } // end while
        } catch (IOException ioe) {
            throw new SIPParseException("Error parsing SIP message: " + ioe.getMessage());
        } // end try

        msg.setHeaders(headers.toString());
        msg.setContents(contents.toString());
        debug.append("--------------------------------------------------------\n");
        System.out.println(debug.toString());
        return msg;
    } // end getMessage()
    
    /** put an entire SIP mesage to the connection.  The parser does minimal checking
     * to see that the message has the minimal SIP requirements.  If it does not an
     * exception is thrown.
     *
     * @param mesg an encapsulated SIP message to send
     */
    public void putMessage(SIPMessage mesg) {
        System.out.println("writing message to " + _socket.getInetAddress().getHostAddress() + ", port " + _socket.getPort());
        _writer.print(mesg.toString());
        _writer.flush();
    } // end putMessage()
    
} // end class SIPParser